#include "skiplist_multimap.hpp"

#include <boost/pool/pool_alloc.hpp>
#include <boost/timer.hpp>

#include <iostream>
#include <map>

template <typename K, typename V>
struct slmap {
	typedef skiplist_multimap<K, V, std::less<K>, std::allocator<V> > type;
	typedef typename type::iterator iter;
	typedef typename type::key_type key_type;
	typedef typename type::value_type value_type;
	typedef void tag;
};

template <typename K, typename V>
struct slmap_bfp {
	typedef skiplist_multimap<K, V, std::less<K>, boost::fast_pool_allocator<V, boost::default_user_allocator_new_delete, boost::details::pool::null_mutex> > type;
	typedef typename type::iterator iter;
	typedef typename type::key_type key_type;
	typedef typename type::value_type value_type;
	typedef int tag;
};

template <typename K, typename V>
struct stdmap {
	typedef std::multimap<K, V, std::less<K>, std::allocator<V> > type;
	typedef typename type::iterator iter;
	typedef typename type::key_type key_type;
	typedef typename type::value_type value_type;
	typedef float tag;
};

template <typename K, typename V>
struct stdmap_bfp {
	typedef std::multimap<K, V, std::less<K>, boost::fast_pool_allocator<V> > type;
	typedef typename type::iterator iter;
	typedef typename type::key_type key_type;
	typedef typename type::value_type value_type;
	typedef double tag;
};

template <typename T>
struct timings {
	static double insert, find, iterate, destroy;
};
template <typename T>
double timings<T>::insert = 0;
template <typename T>
double timings<T>::find = 0;
template <typename T>
double timings<T>::iterate = 0;
template <typename T>
double timings<T>::destroy = 0;


template <typename T>
struct accumulate {
	typedef T type;
	static void apply(type& a, const type& b) { a += b; }
};

template <>
struct accumulate<std::string> {
	typedef size_t type;
	static void apply(type& a, const std::string& b) { a += b.size(); }
};

template <typename T>
void gen_random(T* vals, size_t size) {
	for (int i = 0; i < size; i++)
		vals[i] = rand() / (T)1.1;
}

template <>
void gen_random<std::string>(std::string* vals, size_t size) {
	for (int i = 0; i < size; i++) {
		char str[101];
		for (int t = 0; t < 101; t++)
			str[t] = 30 + (rand() % 70);
		str[rand() % 101] = 0;
		vals[i] = str;
	}
}

template <typename map_t, size_t size>
void run_test(int multiples, int insert_rep, int find_reps, int iterate_reps) {
	typename map_t::type m;
	typename map_t::key_type keys[size];
	typename map_t::value_type::second_type vals[size];

	gen_random<typename map_t::key_type>(keys, size);
	gen_random<typename map_t::value_type::second_type>(vals, size);

	double insertion = 0;
	double find = 0;
	double iterate = 0;
	double destruct = 0;
	double ep = 0;
	boost::timer t;

	for (int i = 0; i < insert_rep; i++) {
		ep = t.elapsed();
		for (int k = 0; k < multiples; ++k) {
			for (int j = 0; j < size; j++)
				m.insert(std::make_pair(keys[j], vals[j]));
		}
		insertion += t.elapsed() - ep;
		ep = t.elapsed();
		m.clear();
		destruct += t.elapsed() - ep;
	}

	for (int i = 0; i < multiples; ++i)
		for (int j = 0; j < size; j++)
			m.insert(std::make_pair(keys[j], vals[j]));

	ep = t.elapsed();
	for (int i = 0; i < find_reps; i++) {
		for (int j = 0; j < size; j++)
			m.find(keys[j]);
	}
	find += t.elapsed() - ep;

	ep = t.elapsed();
	typename accumulate<typename map_t::type::key_type>::type val;
	for (int i = 0; i < iterate_reps; i++) {
		for (typename map_t::type::iterator iter = m.begin(); iter != m.end(); ++iter) {
			accumulate<typename map_t::type::key_type>::apply(val, iter->first);
		}
	}
	iterate += t.elapsed() - ep;

	std::cout << insertion << " " << find << " " << iterate << " " << destruct << std::endl;
	typedef timings<typename map_t::tag> timings_t;
	timings_t::insert += insertion;
	timings_t::find += find;
	timings_t::iterate += iterate;
	timings_t::destroy += destruct;
}

template <typename T, typename K, size_t size>
void test_all(int a, int b, int c, int d) {
	std::cout << "\n #elements: " << size << " insert multiples: " << a << " insert reps: " << b << " find reps: " << c << " iterate reps: " << d << std::endl;
	std::cout << "std::map<std::alloc>  ";
	run_test<stdmap<T, K>, size>(a, b, c, d);
	std::cout << "std::map<boost::fpa>  ";
	run_test<stdmap_bfp<T, K>, size>(a, b, c, d);
	std::cout << "skiplist<std::alloc>  ";
	run_test<slmap<T, K>, size>(a, b, c, d);
	std::cout << "skiplist<boost::fpa>  ";
	run_test<slmap_bfp<T, K>, size>(a, b, c, d);
}

int main() {
	std::cout << "\n\nmap<int, string>" << std::endl;
	test_all<int, std::string, 10>(1, 500000, 3000000, 5000000);
	test_all<int, std::string, 3000>(300, 1, 100, 20);
	test_all<int, std::string, 500000>(1, 1, 1, 5);
	std::cout << "\n\nmap<int, int>" << std::endl;
	test_all<int, int, 10>(1, 1000000, 1000000, 10000000);
	test_all<int, int, 3000>(500, 1, 100, 5);
	test_all<int, int, 1000000>(1, 1, 1, 5);
	std::cout << "\n\nmap<string, string>" << std::endl;
	test_all<std::string, std::string, 5>(1, 1000000, 1000000, 30000000);
	test_all<std::string, std::string, 500>(1000, 1, 1000, 20);
	test_all<std::string, std::string, 300000>(1, 1, 1, 50);
	std::cout << "\n\nmap<double, char>" << std::endl;
	test_all<double, char, 10>(1, 1000000, 2000000, 10000000);
	test_all<double, char, 1000>(10, 300, 3000, 5000);
	test_all<double, char, 500000>(1, 2, 1, 10);

	{
		typedef timings<void> timing_t;
		std::cout << "slmap timings     -- insertion: " << timing_t::insert << " find: " << timing_t::find << " iterate: " << timing_t::iterate << " destruction: " << timing_t::destroy << std::endl;
	}
	{
		typedef timings<int> timing_t;
		std::cout << "slmapbfp timings  -- insertion: " << timing_t::insert << " find: " << timing_t::find << " iterate: " << timing_t::iterate << " destruction: " << timing_t::destroy << std::endl;
	}
	{
		typedef timings<float> timing_t;
		std::cout << "stdmap timings    -- insertion: " << timing_t::insert << " find: " << timing_t::find << " iterate: " << timing_t::iterate << " destruction: " << timing_t::destroy << std::endl;
	}
	{
		typedef timings<double> timing_t;
		std::cout << "stdmapbfp timings -- insertion: " << timing_t::insert << " find: " << timing_t::find << " iterate: " << timing_t::iterate << " destruction: " << timing_t::destroy << std::endl;
	}
	return 0;
}
